﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace Reflector
{
    public class DelegateToReflector<TTarget, TValue> : ISetValue
    {
        Action<TTarget, TValue> _setter;

        public DelegateToReflector(PropertyInfo propertyInfo)
        {
            if (propertyInfo == null)
                throw new ArgumentNullException("propertyInfo");

            if (propertyInfo.CanWrite == false)
                throw new NotSupportedException("属性不支持写操作。");

            MethodInfo m = propertyInfo.GetSetMethod(true);
            _setter = (Action<TTarget, TValue>)Delegate.CreateDelegate(typeof(Action<TTarget, TValue>), null, m);
        }

        public void Set(object target, object value)
        {
            _setter((TTarget)target, (TValue)value);
        }

        public static void Test()
        {
            Person person = new Person();
            PropertyInfo propInfo = typeof(Person).GetProperty("Name");

            //只能创建强类型委托
            //Action<Person, string> setter = (Action<Person, string>)Delegate.CreateDelegate(
            //    typeof(Action<Person, string>), null, propInfo.GetSetMethod());

            //放开这段代码，你会发现编译错误
            Action<object, object> setter = (Action<object, object>)Delegate.CreateDelegate(
                typeof(Action<object, object>), null, propInfo.GetSetMethod());

            setter(person, "Test");

            Console.WriteLine(person.Name);
        }
    }

    public class GetterSetterFactory
    {
        public static ISetValue CreatePropertySetterWrapper(PropertyInfo propertyInfo)
        {
            if (propertyInfo == null)
                throw new ArgumentNullException("propertyInfo");
            if (propertyInfo.CanWrite == false)
                throw new NotSupportedException("属性不支持写操作。");

            MethodInfo mi = propertyInfo.GetSetMethod(true);

            if (mi.GetParameters().Length > 1)
                throw new NotSupportedException("不支持构造索引器属性的委托。");

            Type instanceType = typeof(DelegateToReflector<,>).MakeGenericType(propertyInfo.DeclaringType, propertyInfo.PropertyType);
            return (ISetValue)Activator.CreateInstance(instanceType, propertyInfo);
        }
    }

    public interface ISetValue
    {
        void Set(object target, object val);
    }
}
