﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;
using System.Linq.Expressions;

namespace Reflector
{
    public class ExpressionToReflector
    {
        public static Action<object, object> InitSetter(PropertyInfo propertyInfo)
        {
            //((T)instance).Property = ((V)parm)

            //定义两个参数：instance（属性的拥有者），parm（要设置的值）
            var instance = Expression.Parameter(typeof(object), "instance");
            var parm = Expression.Parameter(typeof(object), "parm");

            //判断方法是否是静态的，若非静态方法，需要把instance转换为对应的类型
            var instanceCast = propertyInfo.GetSetMethod().IsStatic ? null :
                Expression.Convert(instance, propertyInfo.ReflectedType);

            //参数类型转换（(V)parm）
            var parmCast = Expression.Convert(parm, propertyInfo.PropertyType);

            //((T)instance).Property
            var property = Expression.Property(instanceCast, propertyInfo);

            //属性赋值操作  ((T)instance).Property = ((V)parm)
            var setProperty = Expression.Assign(property, parmCast);

            //定义委托
            var lambda = Expression.Lambda<Action<object, object>>(setProperty, instance, parm);
            //创建委托
            return lambda.Compile();
        }
    }
}
