﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Reflection;
using System.Text;
using System.Threading.Tasks;

namespace Reflector
{
    class Program
    {
        static void Main(string[] args)
        {
            //TestEmitToReflector();

            //DelegateToReflector<Person,string>.Test();

            //TestDelegateToReflector();

            TestExpressionToReflector();
        }

        static void TestEmitToReflector()
        {
            int count = 1000000;

            Person testObj = new Person();
            PropertyInfo propInfo = typeof(Person).GetProperty("Name");

            Console.Write("直接访问花费时间：       ");
            Stopwatch watch1 = Stopwatch.StartNew();

            for (int i = 0; i < count; i++)
                testObj.Name = "Test";

            watch1.Stop();
            Console.WriteLine(watch1.Elapsed.ToString());

            EmitToReflector.SetValueDelegate setter2 = EmitToReflector.CreatePropertySetter(propInfo);
            Console.Write("EmitSet花费时间：        ");
            Stopwatch watch2 = Stopwatch.StartNew();

            for (int i = 0; i < count; i++)
                setter2(testObj, "Test");

            watch2.Stop();
            Console.WriteLine(watch2.Elapsed.ToString());


            Console.Write("纯反射花费时间：　       ");
            Stopwatch watch3 = Stopwatch.StartNew();

            for (int i = 0; i < count; i++)
                propInfo.SetValue(testObj, "Test", null);

            watch3.Stop();
            Console.WriteLine(watch3.Elapsed.ToString());


            Console.WriteLine("-------------------");
            Console.WriteLine("{0} / {1} = {2}",
                watch3.Elapsed.ToString(),
                watch1.Elapsed.ToString(),
                watch3.Elapsed.TotalMilliseconds / watch1.Elapsed.TotalMilliseconds);

            Console.WriteLine("{0} / {1} = {2}",
                watch3.Elapsed.ToString(),
                watch2.Elapsed.ToString(),
                watch3.Elapsed.TotalMilliseconds / watch2.Elapsed.TotalMilliseconds);

            Console.WriteLine("{0} / {1} = {2}",
                watch2.Elapsed.ToString(),
                watch1.Elapsed.ToString(),
                watch2.Elapsed.TotalMilliseconds / watch1.Elapsed.TotalMilliseconds);
        }

        static void TestDelegateToReflector()
        {
            int count = 1000000;

            Person testObj = new Person();
            PropertyInfo propInfo = typeof(Person).GetProperty("Name");

            Console.Write("泛型委托花费时间：       ");
            DelegateToReflector<Person, string> setter3 = new DelegateToReflector<Person, string>(propInfo);
            Stopwatch watch4 = Stopwatch.StartNew();

            for (int i = 0; i < count; i++)
                setter3.Set(testObj, "Test");

            watch4.Stop();
            Console.WriteLine(watch4.Elapsed.ToString());


            Console.Write("通用接口花费时间：       ");
            var setter4 = GetterSetterFactory.CreatePropertySetterWrapper(propInfo);
            Stopwatch watch5 = Stopwatch.StartNew();

            for (int i = 0; i < count; i++)
                setter4.Set(testObj, "Test");

            watch5.Stop();
            Console.WriteLine(watch5.Elapsed.ToString());
        }

        static void TestExpressionToReflector()
        {
            int count = 1000000;

            Person testObj = new Person();
            PropertyInfo propInfo = typeof(Person).GetProperty("Name");

            Console.Write("泛型委托花费时间：       ");
            DelegateToReflector<Person, string> setter3 = new DelegateToReflector<Person, string>(propInfo);
            Stopwatch watch4 = Stopwatch.StartNew();

            for (int i = 0; i < count; i++)
                setter3.Set(testObj, "Test");

            watch4.Stop();
            Console.WriteLine(watch4.Elapsed.ToString());


            Console.Write("通用接口花费时间：       ");
            var setter4 = ExpressionToReflector.InitSetter(propInfo);
            Stopwatch watch5 = Stopwatch.StartNew();

            for (int i = 0; i < count; i++)
                setter4(testObj, "Test");

            watch5.Stop();
            Console.WriteLine(watch5.Elapsed.ToString());
        }
    }
}
